
/*	OSS (open sound system) sound driver

***	Copyright:	1999 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	05.Jan.99:	start on this file			kio
*/


#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <errno.h>

#include "kio.h"

#include "options.h"
#include "audio.h"

#ifndef LOG
#define LOG 1
#endif
#ifndef SAFE
#define SAFE 2
#endif


/* ===============================================================
		audio output
		Hardware dependant part
		via OSS (open audio system)
=============================================================== */


/* ----	create dsp instance --------------------------------------
*/
Dsp::Dsp ( )
{
	XLogLine("Dsp::Dsp()");
	MemClear ( this, sizeof(Dsp) );
	clock = SAMPLECLOCK;
}


/* ----	destroy instance -----------------------------------------
*/
Dsp::~Dsp ( )
{
	XLogLine(">Dsp::~Dsp()");
	Close();
	XLogExit();
}


/* ----	open dsp ---------------------------------------------------
	return: errno == ok or error
*/
OSErr Dsp::Open ( )
{
	XLogLine ( ">Dsp::Open()" );	

	#define SAMPLEFORMAT	(sizeof(Sample)==1 ? AFMT_U8 : AFMT_S16_LE)
	#define	SAMPLEBITS 	LOG2(sizeof(Sample))
	XXTRAP((1<<SAMPLEBITS)!=sizeof(Sample));

	int fragarg = (CHUNKS<<16) + (SAMPLEBITS+CHUNKBITS);
	int fmtarg  = SAMPLEFORMAT;

	errno = ok;

	if ( On() ) goto x;

	clock = SAMPLECLOCK;

// open sound output device
	fd = open ( DEV_AUDIO, O_WRONLY, 0);
	if ( DEV_AUDIO_AUX+0 && fd==-1 ) 
	fd = open ( DEV_AUDIO_AUX+0, O_WRONLY, 0);	
	TRAP(fd==0);

	if (fd==-1)
	{
		Log ( "failed to open \"%s\": %i = %s", DEV_AUDIO, errno, strerror(errno) );
		fd = 0;
	}

// set dsp buffer chunks: number and size
	else if ( ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &fragarg))
	{
		Log ( "failed to setup sound buffer parameters: %s", strerror(errno) );
		close(fd);
		fd = 0;
	}

// set dsp sample size
// ioctl modiefies fmtarg!
	else if ( ioctl(fd, SNDCTL_DSP_SETFMT, &fmtarg)==-1 || fmtarg != SAMPLEFORMAT )  
	{
		Log ( "failed to set sample size: %s", strerror(errno) );
		close(fd);
		fd = 0;
	}

// set dsp sample clock
// ioctl() modifies clock!
	else if ( ioctl(fd, SNDCTL_DSP_SPEED, &clock)==-1 )
	{
		Log ( "failed to set sampling frequency: %s", strerror(errno) );
		close(fd); 
		fd = 0;
	}

	else
	{
// allocate dsp buffer
		OMEM ( buffer = bufferptr = new Sample[CHUNKSIZE] );
		bufferend = buffer + CHUNKSIZE;
		errno = ok;
	}

x:	XLogExit();
	return errno;
}


/* ----	close dsp --------------------------------------------------
*/
void Dsp::Close ( )
{
	XLogLine(">Dsp::Close()");

	delete[] buffer;
	if (fd) close(fd);
	MemClear(this,sizeof(Dsp));
	// fd=0;
	// buffer = bufferptr = bufferend = 0;
	// state = dsp_off;

	XLogExit();
}


/* ----	querry state of sample buffer ----------------------------------
	returned value:
		dsp_off, dsp_busy	=> off
		dsp_empty, dsp_low	=> on, please hurry up
		dsp_fine		=> on
		dsp_high, dsp_full	=> on, please slow down
*/
int Dsp::State ( )
{
	XXXLogLine("dsp::State()");

	if (fd)
	{
		audio_buf_info info;
		long n;
		TRAP ( ioctl(fd, SNDCTL_DSP_GETOSPACE, &info) == -1 );

		n = info.bytes/sizeof(Sample) + (bufferend-bufferptr);
		n = n*10 / ((CHUNKS+1)*CHUNKSIZE);

		switch ( n )
		{
		case 0:
		case 1:	return dsp_full;
		case 2:	
		case 3:	return dsp_high;	
		case 4:
		case 5:	return dsp_fine;
		case 6:
		case 7:
		case 8:	return dsp_low;
		case 9:	
		case 10:return dsp_empty;
		}
	}

	// if (???) return dsp_busy;
	return dsp_off;
}


/* ----	actually write buffer to /dev/dsp ------------------------------
*/
OSErr Dsp::Write ( )
{
/* 	the system clock:
	either the dsp (real world) or the emulated cpu (virtual world) runs faster, 
	because it's very hard to make them run exactly at the same speed.

	if the dsp runs faster, it's sound queue tends to become empty (underrun).
	then the virtual machine must hurry up, e.g. by executing an extra frame.

	if the dsp runs slower, it's sound queue tends to become full. 
	then the virtual machine must wait until a sound chunk becomes free.

	write(...) to /dev/dsp may return error 4 (interrupted system call) 
	when an interrupt occurs while we are blocked by the oss audio device.
	this seems to result in audible effects. 
	therefore we test before writing whether there is enough space in the
	oss sound buffer for our call to write(...)
*/
	XXLogLine ( ">Dsp::Write()" );

	audio_buf_info info;
	int     n;
	Sample *bp;

	errno = ok;

	if (!On())
	{
		//if (State()==dsp_busy) {}
		errno = ENXIO;
		goto x;
	}

	if (XXXLOG) for (int i=0;i<CHUNKSIZE;i++) Log("%02X",(int)buffer[i]);

	TRAP ( ioctl(fd, SNDCTL_DSP_GETOSPACE, &info) == -1 );
	if (info.fragments<1) 
	{ 
	 	XLog("dsp pause");
		do
		{	pause();
			TRAP ( ioctl(fd, SNDCTL_DSP_GETOSPACE, &info) == -1 );
		}
		while (info.fragments<1);
	}

	n  = CHUNKSIZE*sizeof(Sample);
	bp = buffer;
	if ( write(fd,bp,n) == -1 )
	{
		Log("could not feed sound device: %i = %s", errno, strerror(errno));
		Log("switching off sound");
		Close();
	}

x:	XXLogExit();

	bufferptr = buffer;
	return errno;
}




/* ----	write sample buffer to dsp -----------------------------------------
*/
OSErr Dsp::Write ( Sample *data, int count )
{
	XXXLogLine ( ">Dsp::Write($%p,%i)", data,count );

	if (!On()) { errno=ENXIO; goto x; }

	while ( bufferptr+count >= bufferend )
	{
		long n = bufferend-bufferptr;
		MemCopy ( data, bufferptr, n*sizeof(Sample) );
		data  += n;
		count -= n;
		if ( Write()!=ok ) goto x;
	}

	memcpy ( bufferptr, data, count*sizeof(Sample) );
	bufferptr += count;
	errno = ok;

x:	XXXLogExit();
	return errno;
}





