/*
***	Copyright:	1998-1999 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#include <stdarg.h>
#include <string.h>
#include "kio.h"


// ==========================================================================
// ============================== error messages ============================
// ==========================================================================


char* ErrorText ( OSErr err )
{
	if ( (Int)(-1-err<errorTextSize) )	return errorText[-1-err];
	else					return strerror(err);
}


/* ----	special function supplied for the 'TRAP()' macro (quick&dirty error checker)
	the application must provide extern errorText[] and extern errorTextSize
	which must at least contain texts for general error -1 and internal error -2.
*/
void _trap_error_ ( char*file, long line, OSErr err )
{
	if (err)
	{
		Log ( "Sorry, a fatal error occured:" );
		Log ( "File: \"%s\"; in line: %i", file, line );
		Log ( ErrorText(err) );
		abort();
	}
}


// ===========================================================================
// ============================= logging to stdout ===========================
// ===========================================================================


static const int WIDTH   = 80;		// desired max. text width
static int   	 indent  = 0;		// current indentation
static char 	 space[] = "                                ";
static int  	 column  = 0;		// current text column


void LogNL ( )
{
	if (column) printf("\n");
	column = 0;
}


void Log ( const char* fmt, ... )
{
	if (!fmt||!*fmt) return;

	char text[500];
	char buff[500];
	va_list va;
	char *t,*b,c;

	va_start(va,fmt);

	t = text;
	
	while ( *fmt )
	{
	// copy plain text from format string
		while ( *fmt && *fmt!='%' && t<text+sizeof(text) )  *t++ = *fmt++;
	
	// test for end of format string
		if (*fmt!='%') break;

	// copy format to buffer and expand and append it to text
		b=buff;
		*b++ = *fmt++;			// '%'
			
		if (*fmt=='%') {fmt++;continue;}// "%%" := '%'
			
		char sz = 0;			// argument size specifier (if found)

		int  nn = text+sizeof(text)-t;	// remaining free space in text

		for (;;)
		{	
			c = *b++ = *fmt++;	// copy format character
			*b = 0;			// preset string end
				
			switch ( c )
			{
			case '0':case '1':case '2':case '3':case '4':case '5':case '6':
			case '7':case '8':case '9':case '#':case '-':case '+':
			case ' ':case '.':	// precision, filler
				continue;
			case 'h':		// diouxX := short;     n     := *short
			case 'l':		// diouxX := long;      n     := *long
			case 'q':case 'L':	// diouxX := long long; eEfgG := long double
			case 'Z':		// diouxX := size_t
				sz = c;
				continue;
			case '*':		// an int argument supplies precision
				NIMP();
			case 'd':case 'i':case 'o':case 'u':case 'x':case 'X':	// int
				if      (sz== 0 )	t += snprintf ( t, nn, buff, va_arg(va,int) );
				else if (sz=='h')	t += snprintf ( t, nn, buff, va_arg(va,short) );
				else if (sz=='l')	t += snprintf ( t, nn, buff, va_arg(va,long) );
				else if (sz=='Z')	t += snprintf ( t, nn, buff, va_arg(va,size_t) );
				else 			t += snprintf ( t, nn, buff, va_arg(va,long long) );
				break;
			case 'e':case 'E':case 'f':case 'g':case 'G':	// double
				if	( sz== 0 )	t += snprintf ( t, nn, buff, va_arg(va,double) );
				else 			t += snprintf ( t, nn, buff, va_arg(va,long double) );
				break;
			case 'c':		// character
				t += snprintf ( t, nn, buff, va_arg(va,int) );
				break;
			case 's':		// string
				t += snprintf ( t, nn, buff, va_arg(va,char*) );
				break;
			case 'p':		// pointer
				t += snprintf ( t, nn, buff, va_arg(va,void*) );
				break;
			case 'n':		// store number of char written so far
				NIMP();
			default:		// no valid format: e.g. '%' or '\0' or s.th. else
				IERR();
			}

			break;			// leave infinite loop
		}
	}

	*t = 0;

// if printing starts in a new line, '<' and '>' control indentation:

	if (column==0) switch ( *text )
	{
	case '>':
		Log ( text+1 );
		if (indent<=(int)sizeof(space)-1-2) indent+=2;
		return;
	case '<':
		if (indent>=2) indent-=2;
		Log ( text+1 );
		return;
	}

// at last: print it

	if (column+strlen(text)>=(int)WIDTH) LogNL();
	int n = column==0 ? indent : 1;
	column += printf ( "%s%s", space+sizeof(space)-1-n, text );

	
	va_end(va);
}









